           (*-------------------------------------------*
            |                October 2007               |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_F_law_decompo_v4_1
imports CSP_F_law_decompo CSP_syntax_v4_1
begin

(*****************************************************************

         1. 

 *****************************************************************)

(* ----------------------------------------------------------- *

          decomposition for sending and receiving

 * ----------------------------------------------------------- *)

(*********************************************************
                        Act_prefix mono
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

(** mono **)

lemma cspF_Send_prefix_mono:
  "[| a = b ; P <=F[M1,M2] Q |] ==> a ! v -> P <=F[M1,M2] b ! v -> Q"
apply (simp add: csp_prefix_ss_def)
apply (simp add: cspF_decompo)
done

lemma cspF_Rec_prefix_mono:
  "[| inj a; a = b ; X = Y ; 
      !! x. x:Y ==> Pf x <=F[M1,M2] Qf x |]
    ==> a ? x:X -> Pf x <=F[M1,M2] b ? x:Y -> Qf x"
apply (simp add: csp_prefix_ss_def)
apply (rule cspF_Ext_pre_choice_mono)
apply (simp)
apply (simp add: image_iff)
apply (erule bexE)
apply (simp)
done

lemma cspF_Int_pre_choice_mono:
  "[| X = Y ; 
      !! x. x:Y ==> Pf x <=F[M1,M2] Qf x |]
    ==> ! :X -> Pf <=F[M1,M2] ! :Y -> Qf"
apply (simp add: csp_prefix_ss_def)
apply (simp add: cspF_decompo)
done

lemma cspF_Nondet_send_prefix_mono:
  "[| inj a; a = b ; X = Y ; 
      !! x. x:Y ==> Pf x <=F[M1,M2] Qf x |]
    ==> a !? x:X -> Pf x <=F[M1,M2] b !? x:Y -> Qf x"
apply (simp add: csp_prefix_ss_def)
apply (rule cspF_mono)
apply (simp)
apply (rule cspF_mono)
apply (simp)
apply (simp add: image_iff)
apply (erule bexE)
apply (simp)
done

(** cong **)

lemma cspF_Send_prefix_cong:
  "[| a = b ; P =F[M1,M2] Q |] ==> a ! v -> P =F[M1,M2] b ! v -> Q"
apply (simp add: csp_prefix_ss_def)
apply (simp add: cspF_decompo)
done

lemma cspF_Rec_prefix_cong:
  "[| inj a; a = b ; X = Y ; 
      !! x. x:Y ==> Pf x =F[M1,M2] Qf x |]
    ==> a ? x:X -> Pf x =F[M1,M2] b ? x:Y -> Qf x"
by (simp add: cspF_eq_ref_iff cspF_Rec_prefix_mono)

lemma cspF_Int_pre_choice_cong:
  "[| X = Y ; 
      !! x. x:Y ==> Pf x =F[M1,M2] Qf x |]
    ==> ! :X -> Pf =F[M1,M2] ! :Y -> Qf"
apply (simp add: csp_prefix_ss_def)
apply (simp add: cspF_decompo)
done

lemma cspF_Nondet_send_prefix_cong:
  "[| inj a; a = b ; X = Y ; 
      !! x. x:Y ==> Pf x =F[M1,M2] Qf x |]
    ==> a !? x:X -> Pf x =F[M1,M2] b !? x:Y -> Qf x"
by (simp add: cspF_eq_ref_iff cspF_Nondet_send_prefix_mono)

lemmas cspF_prefix_ss_mono = cspF_Send_prefix_mono
                             cspF_Rec_prefix_mono
                             cspF_Int_pre_choice_mono
                             cspF_Nondet_send_prefix_mono

lemmas cspF_prefx_ss_cong = cspF_Send_prefix_cong
                            cspF_Rec_prefix_cong
                            cspF_Int_pre_choice_cong
                            cspF_Nondet_send_prefix_cong

(*********************************************************
             Rep_internal_choice for UNIV 
              this is useful for tactic
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

(* mono_UNIV *)

lemma cspF_Rep_int_choice_mono_UNIV_nat:
  "[| !! n. Pf n <=F[M1,M2] Qf n |]
      ==> !nat n .. Pf n <=F[M1,M2] !nat n .. Qf n"
by (simp add: cspF_Rep_int_choice_mono)

lemma cspF_Rep_int_choice_mono_UNIV_set:
  "[| !! X. Pf X <=F[M1,M2] Qf X |]
      ==> !set X .. Pf X <=F[M1,M2] !set X .. Qf X"
by (simp add: cspF_Rep_int_choice_mono)

lemma cspF_Rep_int_choice_mono_UNIV_com:
  "[| !! x. Pf x <=F[M1,M2] Qf x |]
      ==> ! x .. Pf x <=F[M1,M2] ! x .. Qf x"
by (simp add: cspF_Rep_int_choice_mono)

lemma cspF_Rep_int_choice_mono_UNIV_f:
  "[| inj f ; !! x. Pf x <=F[M1,M2] Qf x |]
      ==> !<f> x .. Pf x <=F[M1,M2] !<f> x .. Qf x"
by (simp add: cspF_Rep_int_choice_mono)

lemmas cspF_Rep_int_choice_mono_UNIV = 
      cspF_Rep_int_choice_mono_UNIV_nat
      cspF_Rep_int_choice_mono_UNIV_set
      cspF_Rep_int_choice_mono_UNIV_com
      cspF_Rep_int_choice_mono_UNIV_f

(* cong *)

lemma cspF_Rep_int_choice_cong_UNIV_nat:
  "[| !! n. Pf n =F[M1,M2] Qf n |]
      ==> !nat n .. Pf n =F[M1,M2] !nat n .. Qf n"
by (simp add: cspF_Rep_int_choice_cong)

lemma cspF_Rep_int_choice_cong_UNIV_set:
  "[| !! X. Pf X =F[M1,M2] Qf X |]
      ==> !set X .. Pf X =F[M1,M2] !set X .. Qf X"
by (simp add: cspF_Rep_int_choice_cong)

lemma cspF_Rep_int_choice_cong_UNIV_com:
  "[| !! x. Pf x =F[M1,M2] Qf x |]
      ==> ! x .. Pf x =F[M1,M2] ! x .. Qf x"
by (simp add: cspF_Rep_int_choice_cong)

lemma cspF_Rep_int_choice_cong_UNIV_f:
  "[| inj f ; !! x. Pf x =F[M1,M2] Qf x |]
      ==> !<f> x .. Pf x =F[M1,M2] !<f> x .. Qf x"
by (simp add: cspF_Rep_int_choice_cong)

lemmas cspF_Rep_int_choice_cong_UNIV = 
      cspF_Rep_int_choice_cong_UNIV_nat
      cspF_Rep_int_choice_cong_UNIV_set
      cspF_Rep_int_choice_cong_UNIV_com
      cspF_Rep_int_choice_cong_UNIV_f

end
